from __future__ import division

import sys
from math import trunc, exp, log, ceil, floor
import matplotlib.pyplot as plt
import matplotlib.axis as axis
import matplotlib.ticker as mtick
from matplotlib import rc

import random

from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import schedcat.model.mcx_tasks as ts
import numpy as np
from matplotlib import cm

from schedcat.model.LLCLayout import *

def uniform_int(minval, maxval):
    "Create a function that draws ints uniformly from {minval, ..., maxval}"
    def _draw():
        return random.randint(minval, maxval)
    return _draw

def uniform(minval, maxval):
    "Create a function that draws floats uniformly from [minval, maxval]"
    def _draw():
        return random.uniform(minval, maxval)
    return _draw

def log_uniform(minval, maxval):
    "Create a function that draws floats log-uniformly from [minval, maxval]"
    def _draw():
        return exp(random.uniform(log(minval), log(maxval))) 
    return _draw

def log_uniform_int(minval, maxval):
    "Create a function that draws ints log-uniformly from {minval, ..., maxval}"
    draw_float = log_uniform(minval, maxval + 1)
    def _draw():
        val = int(draw_float())
        val = max(minval, val)
        val = min(maxval, val)
        return val
    return _draw

def uniform_choice(choices):
    "Create a function that draws uniformly elements from choices"
    selector = uniform_int(0, len(choices) - 1)
    def _draw():
        return choices[selector()]
    return _draw

def geometric(p, include_zero = False):
    if include_zero:
        rounder = floor
    else:
        rounder = ceil
    def _draw():
        return rounder(log(1-random.random()) / log(1-p))
    return _draw

def truncate(minval, maxval):
    def _limit(fun):
        def _f(*args, **kargs):
            val = fun(*args, **kargs)
            return min(maxval, max(minval, val))
        return _f
    return _limit

def redraw(minval, maxval):
    def _redraw(dist):
        def _f(*args, **kargs):
            in_range = False
            while not in_range:
                val = dist(*args, **kargs)
                in_range = minval <= val <= maxval
            return val
        return _f
    return _redraw

def exponential(minval, maxval, mean, limiter=redraw):
    """Create a function that draws floats from an exponential
    distribution with expected value 'mean'. If a drawn value is less
    than minval or greater than maxval, then either another value is
    drawn (if limiter=redraw) or the drawn value is set to minval or
    maxval (if limiter=truncate)."""
    def _draw():
        return random.expovariate(1.0 / mean)
    return limiter(minval, maxval)(_draw)

def multimodal(weighted_distributions):
    """Create a function that draws values from several distributions
    with probability according to the given weights in a list of
    (distribution, weight) pairs."""
    total_weight = sum([w for (d, w) in weighted_distributions])
    selector = uniform(0, total_weight)
    def _draw():
        x = selector()
        wsum = 0
        for (d, w) in weighted_distributions:
            wsum += w
            if wsum >= x:
                return d()
        assert False # should never drop off
    return _draw

def restrict(bounds, wMin, wMax):
    interval = bounds[1]-bounds[0]
    return [bounds[0]+wMin*interval, bounds[0]+wMax*interval]

def plot(cost, period, atlevel, file):
    print "plotting"
    fig = plt.figure()
    
    X = np.arange(0, 17, 1)
    Y = np.arange(0, 17, 1)
    X, Y = np.meshgrid(X, Y)
    Z = []
    
    for s in range(17):
        Z.append([])
        for w in range(17):
            Z[s].append(cost[atlevel][s][w])
    
    print X
    print Y
    print Z
    
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.plot_surface(X, Y, Z, rstride=1, cstride=1)
    
    ax.set_zlim(0, 7000)
    ax.view_init(40, 45)
    plt.savefig("../PETS/" + str(file) + ".png", bbox_inches='tight', pad_inches=0.1)

    plt.close()
    
    
    print "done"


class Custom:
    pass
    
class TaskGeneratorMCX(object):
    """Sporadic task generator"""

    def __init__(self, period, util, util_dist, mem_util, avg_percent, a_inflation, pfuncs, mode_count, mode_olap, crit_change, util_var, pet_sens, dram_dist, stat_dist, maxWays = 16, maxSets = 16, deadline=lambda x, y: y, ):
        """Creates TaskGenerator based on a given a period and
        utilization distributions."""
        self.period    = period
        self.util      = util
        self.deadline  = deadline
        self.mW = maxWays
        self.mS = maxSets
        self.util_dist = util_dist
        self.mem_util = mem_util
        self.P = pfuncs
        self.avg_percent = avg_percent
        self.a_inflation = a_inflation
        
        self.mode_count = mode_count
        self.mode_olap = mode_olap
        self.crit_change = crit_change
        self.util_var = util_var
        self.pet_sens = pet_sens
        self.dram_dist = dram_dist
        self.stat_dist = stat_dist
    
#     def Pv(self, S, W, Psam):
#         return (1-Psam.kmin)*exp(-S*Psam.ks-W*Psam.kw-S*W*Psam.kw)+Psam.kmin

    def Pv(self, sz, Psam):
        return (1-Psam.kmin)*exp(-sz*Psam.ks)+Psam.kmin
        
        
    def Pv(self, sz, Psam):
        return (1-Psam.kmin)*exp(-sz*Psam.ks)+Psam.kmin
    
    def Psample(self, P):
        Psamp = Custom()
        
        Psamp.ks = P.ks()
        Psamp.kw = P.kw()
        Psamp.kp = P.kp()
        Psamp.kmin = P.kmin()
        return Psamp

    def tasks(self, tot_mc2_util, max_tasks=None, max_util=None, squeeze=False,
              time_conversion=trunc, a_inflate = 1.5, level = 'C'):
        """Generate a sequence of tasks until either max_tasks is reached
        or max_util is reached. If max_util would be exceeded and squeeze is
        true, then the last-generated task's utilization is scaled to exactly
        match max_util. Otherwise, the last-generated task is discarded.
        time_conversion is used to convert the generated (non-integral) values
        into integral task parameters.
        """
        count = 0
        usum  = 0
        while ((max_tasks is None or count < max_tasks) and
               (max_util is None  or usum  < max_util)):
            period   = self.period[level]()
            util     = self.util[level]()
            cost     = period * util
            deadline = self.deadline(cost, period)
            # scale as required
            period   = max(1, int(time_conversion(period)))
            cost     = max(1, int(time_conversion(cost)))
            deadline = max(1, int(time_conversion(deadline)))
            tsk_pet_sens = {'way' : self.pet_sens['way'](), 'color' : self.pet_sens['color']()}
            util = cost / period
            count  += 1
            usum   += util
            if max_util and usum > max_util:
                if squeeze:
                    # make last task fit exactly
                    util -= (usum - max_util)
                    cost = trunc(period * util)
                else:
                    break
            
            if cost > 0:
                wss = self.mem_util()*cost/(1.5*ohc.CPMD) #of 1-way X 1-color LLC area "units" 
                    
                # ----Generate A and B Partition Functions---- #
                PA = self.Psample(self.P[level]['A'])
                               
                scales = {'A': [[self.Pv(min(s*w, wss), PA) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]}
                
                # First set up A
                for i in range(0, self.mW+1):
                    scales['A'][0][i] = 1.0
                for i in range(0, self.mS+1):
                    scales['A'][i][0] = 1.0
                        
                minA = scales['A'][self.mS][self.mW]
                for s in range(0, self.mS+1):
                    for w in range(0, self.mW+1):
                        scales['A'][s][w] /= minA
            
                # Then B
                scales['B'] = [[scales['A'][s][w]/a_inflate for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                costs = dict()
                
                for key in scales.keys():
                    costs[key] = [[scale*cost for scale in scales_s] for scales_s in scales[key]]
                    
                
                backload = tot_mc2_util - costs['A'][self.mW][self.mS]/period
                
                c_percent = self.avg_percent(backload, level)
                
                costs['Ciso'] = [[costs['B'][s][w]*c_percent for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                highCcost = costs['Ciso'][0][0]
                lowCcost = costs['Ciso'][self.mS][self.mW]
                
                lowCshared = lowCcost+random.uniform(0.3, 0.7)*(highCcost-lowCcost)
                
                costs['Cshared'] = [[highCcost-(float(s*w)/float(self.mW*self.mS))*(highCcost-lowCshared) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                ##########
#                 plot(costs, period, 'A', 'a'+str(count))
                
                if not level is 'C':
                    costs['C'] = [[costs['Ciso'][s][w] for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                for crit in ['A', 'B', 'Cshared', 'Ciso']:
                    costs[crit] = [[costs[crit][16][w]+tsk_pet_sens['color']*(costs[crit][s][w]-costs[crit][16][w]) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                    costs[crit] = [[costs[crit][s][16]+tsk_pet_sens['way']  *(costs[crit][s][w]-costs[crit][s][16]) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                    
#                 plot(costs, period, 'A', 'b'+str(count))
#                 for crit in ['A', 'B', 'Cshared', 'Ciso']:
#                     for s in range(17):
#                         for w in range(17):
#                             if s*w > wss:
#                                 if costs[crit][s][w] < costs[crit][0][0]*PA.kmin:
#                                     costs[crit][s][w] = costs[crit][0][0]*PA.kmin
                        
#                 plot(costs, period, 'A', 'c'+str(count))
#                 sys.exit()
                tsk = ts.SporadicTaskMCX(exec_cost = costs, period = period, wss = wss, level = level, deadline = deadline)
                tsk.pt = ceil(self.dram_dist[tsk.level]())
                tsk.wss = min(tsk.pt, tsk.wss)
                tsk.state = min(128*self.stat_dist[tsk.level]()*tsk.pt, 128*tsk.wss)
                tsk.R = []
                tsk.W = []
                tsk.mode_changer = False
                
                tsk.cc_rep = False
                
                yield tsk
            else:
                break


    def pick_B(self):
        return random.random() < 0.2
        
    def make_mode(self, max_util, prev_mode, mode_ndx):            
        a_inf = self.a_inflation()
        utilA_A = self.util_dist['A']()
        utilB_A = self.util_dist['B']()
        utilC_A = self.util_dist['C']()
        normFactor = max_util/(utilA_A+utilB_A+utilC_A)
        utilA_A *= normFactor
        utilB_A *= normFactor
        utilC_A *= normFactor
        
        utilA_A *= 1+self.util_var()
        utilB_A *= 1+self.util_var()
        utilC_A *= 1+self.util_var()
        
        olapA_A = self.mode_olap()*prev_mode.utilA_A
        olapB_A = self.mode_olap()*prev_mode.utilB_A
        olapC_A = self.mode_olap()*prev_mode.utilC_A
        
        mode = Custom()
        
        mode.utilA_A = 0
        mode.utilB_A = 0
        mode.utilC_A = 0
        mode.ts_A = []
        mode.ts_B = []
        mode.ts_C = []
        mode.bp = 0
        
        
        ndxA = 0
        while mode.utilA_A < olapA_A:
            otsk = prev_mode.ts_A[ndxA]
            new_util = mode.utilA_A + (otsk.cost['A'][self.mS][self.mW]/otsk.period)
            if new_util <= olapA_A and new_util <= utilA_A:
                mode.ts_A.append(otsk)
                mode.utilA_A = new_util
                ndxA += 1
            else:
                break
                
        ndxB = 0
        while mode.utilB_A < olapB_A:
            otsk = prev_mode.ts_B[ndxB]
            new_util = mode.utilB_A + (otsk.cost['A'][self.mS][self.mW]/otsk.period)
            if new_util <= olapB_A and new_util <= utilB_A:
                mode.ts_B.append(otsk)
                mode.utilB_A = new_util
                ndxB += 1
            else:
                break
                
        ndxC = 0
        while mode.utilC_A < olapC_A:
            if(ndxB < len(prev_mode.ts_B) and self.pick_B()):
                otsk = prev_mode.ts_B[ndxB]
                new_util = mode.utilC_A + (otsk.cost['A'][self.mS][self.mW]/otsk.period)
                if new_util <= olapC_A and new_util <= utilC_A:
                    rtsk = ts.SporadicTaskMCX(exec_cost = otsk.cost, period = otsk.period, wss = otsk.wss, level = 'C', deadline = otsk.deadline)
                    rtsk.ways = otsk.ways
                    rtsk.sets = otsk.sets
                    rtsk.pt = otsk.pt
                    rtsk.state = otsk.state
                    rtsk.mode_changer = False
                    
                    buff = Custom()
                    buff.size_m = otsk.state
                    buff.tech = 'SBP'
                    otsk.R = [buff]
                    otsk.W = [buff]
                    rtsk.R = [buff]
                    rtsk.W = [buff]
                    rtsk.cc_rep = True
                
                    mode.bp += ceil(buff.size_m/128)
                
                    mode.ts_C.append(rtsk)
                    mode.utilC_A = new_util
                    ndxB += 1
                else:
                    break
            else:
                otsk = prev_mode.ts_C[ndxC]
                otsk.cc_rep = False
                new_util = mode.utilC_A + (otsk.cost['A'][self.mS][self.mW]/otsk.period)
                if new_util <= olapC_A and new_util <= utilC_A:
                    mode.ts_C.append(otsk)
                    mode.utilC_A = new_util
                    ndxC += 1
                else:
                    break
            
        
        mode_max_util = utilA_A+utilB_A+utilC_A
        
        mode.ts_A.extend(self.tasks(squeeze = True, tot_mc2_util = mode_max_util, level = 'A', max_util = utilA_A-mode.utilA_A, a_inflate = a_inf))
        mode.ts_B.extend(self.tasks(squeeze = True, tot_mc2_util = mode_max_util, level = 'B', max_util = utilB_A-mode.utilB_A, a_inflate = a_inf))
        mode.ts_C.extend(self.tasks(squeeze = True, tot_mc2_util = mode_max_util, level = 'C', max_util = utilC_A-mode.utilC_A, a_inflate = a_inf))
        
        mode.utilA_A = utilA_A
        mode.utilB_A = utilB_A
        mode.utilC_A = utilC_A

        for tsk in mode.ts_A:
            tsk.modes.append(mode_ndx)        
        for tsk in mode.ts_B:
            tsk.modes.append(mode_ndx)        
        for tsk in mode.ts_C:
            tsk.modes.append(mode_ndx)
            
        return mode

    def make_task_set(self, max_util):
        prev_mode = Custom()
        prev_mode.tasks = ts.TaskSystemMCX([])
        prev_mode.utilA_A = 0
        prev_mode.utilB_A = 0
        prev_mode.utilC_A = 0
        
        mode_num = self.mode_count()
        
        bp = 0
        alltasks = []
        for i in range(mode_num):
            prev_mode = self.make_mode(max_util, prev_mode, i)
            alltasks.extend(prev_mode.ts_A)
            alltasks.extend(prev_mode.ts_B)
            alltasks.extend(prev_mode.ts_C)
            bp += prev_mode.bp

        t_i = 0
        while t_i < len(alltasks):
            t_j = t_i + 1
            while t_j < len(alltasks):
                if alltasks[t_i] is alltasks[t_j]:
                    alltasks.pop(t_j)
                else:
                    t_j += 1
            t_i += 1
        
        ts_mcx = ts.TaskSystemMCX(alltasks)
        ts_mcx.a_inflate_choice = 1.5
        ts_mcx.mode_count = mode_num
        ts_mcx.bp = bp
        
        
        ########MODE CHANGER########
        
        mc_period = min([t.period for t in ts_mcx.level('A')])
        mc_costs = dict()
        
        mc_costs['C'] = [[90 for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
        mc_costs['Ciso'] = [[90 for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
        mc_costs['Cshared'] = [[90 for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
        mc_costs['B'] = [[200 for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
        mc_costs['A'] = [[1.5*mc_costs['B'][s][w] for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
        
        
        for b_i in range(4):
            tsk = ts.SporadicTaskMCX(exec_cost = mc_costs, period = mc_period, wss = 0, level = 'A', deadline = mc_period)
            tsk.mode_changer = True
            tsk.modes = range(ts_mcx.mode_count)
            tsk.core = b_i
            tsk.pt = 0 #We factor in mode changer pages elsewhere. Leave this 0 at all times.
            tsk.state = 0
            tsk.R = []
            tsk.W = []
            tsk.cc_rep = False
            ts_mcx.append(tsk)
    
        return ts_mcx
